(*
 * Utilities for the IR.
 *
 * ----------------------------------------------------------------
 *
 * @begin[license]
 * Copyright (C) 2000 Jason Hickey, Caltech
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Author: Jason Hickey
 * @email{jyh@cs.caltech.edu}
 * @end[license]
 *)

open Symbol
open Fc_ir_type

(************************************************************************
 * PRINTING
 ************************************************************************)

(*
 * Default tabstop.
 *)
let tabstop = 3

(*
 * Separated list of fields.
 *)
let print_sep_list sep printer l =
   Format.open_hvbox 0;
   ignore (List.fold_left (fun first x ->
                 if not first then
                    begin
                       Format.print_string sep;
                       Format.print_space ()
                    end;
                 Format.open_hvbox tabstop;
                 printer x;
                 Format.close_box ();
                 false) true l);
   Format.close_box ()

(*
 * Separated list of fields.
 *)
let print_pre_sep_list sep printer l =
   Format.open_hvbox 0;
   ignore (List.fold_left (fun first x ->
                 if not first then
                    begin
                       Format.print_space ();
                       Format.print_string sep
                    end;
                 Format.open_hvbox tabstop;
                 printer x;
                 Format.close_box ();
                 false) true l);
   Format.close_box ()

(*
 * Prefixed list of items.
 *)
let print_prefix_list sep printer l =
   Format.open_hvbox 0;
   List.iter (fun x ->
         Format.print_string sep;
         Format.print_space ();
         Format.open_hvbox tabstop;
         printer x;
         Format.close_box ()) l;
   Format.close_box ()


(*
 * Print a type.
 * Use precedences to reduce the
 * amount of parentheses.
 *)
let prec_none   = 0
let prec_fun    = 1
let prec_struct = 2
let prec_array  = 3

let rec print_type pre = function
   TyAny ->
      Format.print_string "'a"
 | TyChar ->
      Format.print_string "char"
 | TyInt ->
      Format.print_string "int"
 | TyFloat ->
      Format.print_string "float"
 | TyArray (ty, i) ->
      print_type prec_array ty;
      (match i with
          Some i ->
             Format.print_string "[";
             Format.print_int i;
             Format.print_string "]"
        | None ->
             Format.print_string "*")
 | TyStruct fields ->
      if pre >= prec_struct then
         Format.print_string "(";
      Format.open_hvbox 0;
      Format.open_hvbox tabstop;
      Format.print_string "struct {";
      List.iter (fun (v, ty) ->
            Format.print_space ();
            print_type prec_none ty;
            Format.print_string " ";
            print_symbol v;
            Format.print_string ";") fields;
      Format.close_box ();
      Format.print_space ();
      Format.print_string "}";
      if pre >= prec_struct then
         Format.print_string ")";
      Format.close_box ()
 | TyFun (ty_vars, ty_res) ->
      if pre >= prec_fun then
         Format.print_string "(";
      Format.open_hvbox tabstop;
      Format.print_string "(";
      Format.open_hvbox 0;
      print_sep_list "," (print_type prec_none) ty_vars;
      Format.print_string ") ->";
      Format.close_box ();
      Format.print_space ();
      print_type prec_fun ty_res;
      if pre > prec_fun then
         Format.print_string ")";
      Format.close_box ()
 | TyId v ->
      print_symbol v

let print_type = print_type prec_none

(*
 * Atoms.
 *)
let print_atom = function
   AtomChar c ->
      Format.print_string ("\'" ^ Char.escaped c ^ "\'")
 | AtomInt i ->
      Format.print_int i
 | AtomFloat x ->
      Format.print_float x
 | AtomNil ty ->
      Format.print_string "nil[";
      print_type ty;
      Format.print_string "]"
 | AtomVar v ->
      print_symbol v

(*
 * Unary operations.
 *)
let print_unop op =
   let s =
      match op with
         UMinusOp -> "-"
       | UNotOp -> "!"
       | UToChar -> "(char) "
       | UToInt -> "(int) "
       | UToFloat -> "(float) "
       | UToFunPointer -> "(fun *) "
       | UToFun -> "(fun) "
   in
      Format.print_string s

(*
 * Binary operations.
 *)
let print_binop op =
   let s =
      match op with
         PlusOp -> " + "
       | MinusOp -> " - "
       | MulOp -> " * "
       | DivOp -> " / "
       | RemOp -> " % "
       | BAndOp -> " & "
       | BOrOp -> " | "
       | LslOp -> " << "
       | LsrOp -> " >> "
       | AsrOp -> " >> "
       | XorOp -> " ^ "
   in
      Format.print_string s

let print_relop op =
   let s =
      match op with
         EqOp -> " == "
       | NEqOp -> " != "
       | LeOp -> " <= "
       | LtOp -> " < "
       | GtOp -> " > "
       | GeOp -> " >= "
   in
      Format.print_string s

(*
 * "let" opener.
 *)
let print_let_open v ty =
   Format.open_hvbox tabstop;
   Format.print_string "let ";
   print_symbol v;
   Format.print_string " :";
   Format.print_space ();
   print_type ty;
   Format.print_string " =";
   Format.print_space ()

let rec print_let_close e =
   Format.print_string " in";
   Format.close_box ();
   Format.print_space ();
   print_expr e

(*
 * Print an expression.
 *)
and print_expr = function
   LetTypes (types, e) ->
      Format.print_string "let type ";
      print_sep_list "and" (fun (v, ty) ->
            print_symbol v;
            Format.print_string " =";
            Format.print_space ();
            print_type ty) types;
      Format.print_string " in";
      Format.print_space ();
      print_expr e

 | LetFuns (funs, e) ->
      Format.print_string "let ";
      print_pre_sep_list "and " print_function funs;
      Format.print_string " in";
      Format.print_space ();
      print_expr e

 | LetAtom (v, ty, a, e) ->
      print_let_open v ty;
      print_atom a;
      print_let_close e

 | LetCopy (v, ty, a, e) ->
      Format.open_hvbox tabstop;
      Format.print_string "let ";
      print_symbol v;
      Format.print_string " :";
      Format.print_space ();
      print_type ty;
      Format.print_string " =";
      Format.print_space ();
      (match a with
          Some a ->
             Format.print_string "copy(";
             print_atom a;
             Format.print_string ")"
        | None ->
             Format.print_string "uninitialized");
      print_let_close e

 | LetUnop (v, ty, op, a, e) ->
      print_let_open v ty;
      print_unop op;
      print_atom a;
      print_let_close e

 | LetBinop (v, ty, op, a1, a2, e) ->
      print_let_open v ty;
      print_atom a1;
      print_binop op;
      print_atom a2;
      print_let_close e

 | LetApply (v, ty, f, args, e) ->
      print_let_open v ty;
      print_symbol f;
      Format.print_string "(";
      print_sep_list "," print_atom args;
      Format.print_string ")";
      print_let_close e

 | TailCall (f, args) ->
      print_symbol f;
      Format.print_string "(";
      print_sep_list "," print_atom args;
      Format.print_string ")"

 | LetExtCall (v, f, ty, args, e) ->
      Format.open_hvbox tabstop;
      Format.print_string "let ";
      print_symbol v;
      Format.print_string " =";
      Format.print_space ();
      Format.print_string "$external(";
      print_symbol f;
      Format.print_string " :";
      Format.print_space ();
      print_type ty;
      print_prefix_list "," print_atom args;
      Format.print_string ")";
      print_let_close e

 | IntCall (ty, f, args) ->
      Format.print_string "$internal(";
      print_symbol f;
      print_prefix_list "," print_atom args;
      Format.print_string ") :";
      Format.print_space ();
      print_type ty

 | Return a ->
      Format.print_string "$return ";
      print_atom a

 | LetString (v, s, e) ->
      Format.print_string "let ";
      print_symbol v;
      Format.print_string " = \"";
      Format.print_string (String.escaped s);
      Format.print_string "\" in";
      Format.print_space ();
      print_expr e

 | IfThenElse (op, a1, a2, e1, e2) ->
      Format.open_hvbox tabstop;
      Format.print_string "if ";
      print_atom a1;
      Format.print_string " ";
      print_relop op;
      Format.print_string " ";
      print_atom a2;
      Format.print_string " then";
      Format.print_space ();
      print_expr e1;
      Format.close_box ();
      Format.print_space ();
      Format.open_hvbox tabstop;
      Format.print_string "else";
      Format.print_space ();
      print_expr e2;
      Format.close_box ()

 | SetVar (v, ty, a, e) ->
      print_symbol v;
      Format.print_string " : ";
      print_type ty;
      Format.print_string " <- ";
      print_atom a;
      Format.print_string ";";
      Format.print_space ();
      print_expr e

 | LetAddrOfVar (v1, ty, v2, e) ->
      print_let_open v1 ty;
      Format.print_string "&";
      print_symbol v2;
      print_let_close e

 | SetSubscript (a1, ty, a2, a3, e) ->
      (match a2 with
          AtomInt 0 ->
             Format.print_string "*";
             print_atom a1
        | _ ->
             print_atom a1;
             Format.print_string "[";
             print_atom a2;
             Format.print_string "]");
      Format.print_string " : ";
      print_type ty;
      Format.print_string " <- ";
      print_atom a3;
      Format.print_string ";";
      Format.print_space ();
      print_expr e

 | LetSubscript (v, ty, a1, a2, e) ->
      print_let_open v ty;
      print_atom a1;
      Format.print_string "[";
      print_atom a2;
      Format.print_string "]";
      print_let_close e

 | LetAddrOfSubscript (v, ty, a1, a2, e) ->
      print_let_open v ty;
      Format.print_string "&";
      print_atom a1;
      Format.print_string "[";
      print_atom a2;
      Format.print_string "]";
      print_let_close e

 | SetProject (a1, l, ty, a2, e) ->
      print_atom a1;
      Format.print_string "->";
      print_symbol l;
      Format.print_string " : ";
      print_type ty;
      Format.print_string " <- ";
      print_atom a2;
      Format.print_string ";";
      Format.print_space ();
      print_expr e

 | LetProject (v, ty, a, l, e) ->
      print_let_open v ty;
      print_atom a;
      Format.print_string "->";
      print_symbol l;
      print_let_close e

 | LetAddrOfProject (v, ty, a, l, e) ->
      print_let_open v ty;
      Format.print_string "&";
      print_atom a;
      Format.print_string "->";
      print_symbol l;
      print_let_close e

 | Memcpy (a1, ty, a2, e) ->
      Format.print_string "memcpy(";
      print_atom a1;
      Format.open_hvbox 2;
      Format.print_string " : ";
      print_type ty;
      Format.close_box ();
      Format.print_string ", ";
      print_atom a2;
      Format.print_string ");";
      Format.print_space ();
      print_expr e

 | LetAlloc (v, ty, e) ->
      print_let_open v (TyArray (ty, None));
      Format.print_string "alloc(";
      Format.open_hvbox 0;
      print_type ty;
      Format.close_box ();
      Format.print_string ")";
      print_let_close e

 | LetClosure (v, ty, f, args, e) ->
      Format.open_hvbox tabstop;
      Format.print_string "let closure (";
      print_symbol v;
      Format.print_string " : ";
      Format.open_hvbox 0;
      print_type ty;
      Format.close_box ();
      Format.print_string ") =";
      Format.print_space ();
      print_symbol f;
      Format.print_string "(";
      print_sep_list "," print_atom args;
      Format.print_string ") in";
      Format.close_box ();
      Format.print_space ();
      print_expr e

and print_function (f, gflag, f_ty, vars, body) =
   Format.print_string "(";
   if gflag then
      Format.print_string "global ";
   print_symbol f;
   Format.print_string " : ";
   print_type f_ty;
   Format.print_string ")(";
   print_sep_list "," print_symbol vars;
   Format.print_string ") =";
   Format.print_space ();
   print_expr body

(*
 * Wrap in a box.
 *)
let print_expr exp =
   Format.open_vbox 0;
   print_expr exp;
   Format.close_box ()

(*
 * Print a program.
 *)
let print_prog
    { prog_data_type = data_type;
      prog_init_sym = init_sym;
      prog_types = types;
      prog_strings = strings;
      prog_funs = funs
    } =
   Format.open_hvbox tabstop;
   Format.print_string "Program:";
   Format.print_space ();

   Format.open_hvbox tabstop;
   Format.print_string "Data segment type:";
   Format.print_space ();
   print_type data_type;
   Format.close_box ();
   Format.print_space ();

   Format.print_string "Init function: ";
   print_symbol init_sym;
   Format.print_space ();

   Format.open_hvbox tabstop;
   Format.print_string "Types:";
   List.iter (fun (v, ty) ->
         Format.print_space ();
         print_symbol v;
         Format.print_string " = ";
         print_type ty) types;
   Format.close_box ();
   Format.print_space ();

   Format.open_hvbox tabstop;
   Format.print_string "Strings:";
   List.iter (fun (v, s) ->
         Format.print_space ();
         print_symbol v;
         Format.print_string (" = \"" ^ String.escaped s ^ "\"")) strings;
   Format.close_box ();
   Format.print_space ();

   Format.open_hvbox tabstop;
   Format.print_string "Functions:";
   List.iter (fun fund ->
         Format.print_space ();
         Format.open_hvbox tabstop;
         print_function fund;
         Format.close_box ()) funs;
   Format.close_box ()

(*!
 * @docoff
 *
 * -*-
 * Local Variables:
 * Caml-master: "compile"
 * End:
 * -*-
 *)
